import {PostToConnectionCommand} from '@aws-sdk/client-apigatewaymanagementapi';
import {AnyTRPCRouter, callTRPCProcedure, inferRouterContext} from '@trpc/server';
import {APIGatewayProxyEvent, APIGatewayProxyResult, Context} from 'aws-lambda';
import {getApiGatewayManagementApiClient} from '../getApiGatewayManagementApiClient';
import {getConnection} from '../getConnection';
import {TrpcSubscriptionRequest} from './TrpcSubscriptionRequest';
import {TrpcSubscriptionResponse} from './TrpcSubscriptionResponse';
import {isObservable} from './isObservable';

export const awsLambdaTrpcSubscriptionRequestHandler = <Router extends AnyTRPCRouter>(opts: {
  router: Router
  createContext?: (event: APIGatewayProxyEvent) => inferRouterContext<Router> | Promise<inferRouterContext<Router>>
}) => {
  const {router, createContext} = opts;

  return async (event: APIGatewayProxyEvent, context: Context): Promise<APIGatewayProxyResult> => {
    context.callbackWaitsForEmptyEventLoop = false;
    const connection = getConnection(event);

    if (!event.body || event.body === '[]')
      return {statusCode: 200, body: '{}'};

    const request = (JSON.parse(event.body) as TrpcSubscriptionRequest);

    const sendMessage = async (response: TrpcSubscriptionResponse) => {
      await getApiGatewayManagementApiClient(connection.endpoint).send(
          new PostToConnectionCommand({
            ConnectionId: connection.connectionId,
            Data: JSON.stringify(response),
          }),
      );
    };
    const stopConnection = async (): Promise<void> => {
      await sendMessage({
        id: request.id,
        result: {type: 'stopped'},
      });
    };

    try {
      const messageQueue: TrpcSubscriptionResponse[] = [];

      if (request.method === 'subscription.stop') {
        await stopConnection();
      }

      if (request.method === 'subscription') {
        messageQueue.push({
          id: request.id,
          result: {type: 'started'},
        });

        const procedure = (await callTRPCProcedure({
          procedures: router._def.procedures,
          path: request.params.path,
          type: 'subscription',
          ctx: await createContext?.(event) || {},
          getRawInput: async () => request.params.input,
        }));

        if (!isObservable(procedure)) {
          console.error(`observable`, procedure);
          throw `Procedure is not an observable!`;
        }

        procedure.subscribe({
          next: (data: unknown) => {
            messageQueue.push({
              id: request.id,
              result: {
                type: 'data',
                data: data,
              },
            });
          },
          error: (err) => {
            console.error(err);
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            });
          },
          complete: () => {
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            });
          },
        });

        let stopped = false;
        while (!stopped) {
          if (messageQueue.length === 0)
            await new Promise(resolve => setTimeout(resolve, 10));
          const message = messageQueue.shift();
          if (!message)
            continue;

          await sendMessage(message);

          stopped = message.result.type === 'stopped';
        }
      }
    } catch (err) {
      console.error(err);
      await stopConnection();
    }

    return {
      statusCode: 200,
      body: '{}',
    };
  };
};
